/*******************************************************************************
 * Copyright (c) 2007, 2010 Association for Decentralized Information Management
 * in Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.db.tests.api.request.listening;

import java.util.concurrent.Semaphore;

import org.junit.Test;
import org.simantics.db.AsyncReadGraph;
import org.simantics.db.Session;
import org.simantics.db.common.procedure.adapter.AsyncListenerAdapter;
import org.simantics.db.procedure.AsyncProcedure;
import org.simantics.db.request.AsyncRead;
import org.simantics.db.testing.base.ExistingDatabaseTest;
import org.simantics.utils.DataContainer;

public class DisposedListenerTest extends ExistingDatabaseTest {

    /**
     * Performs an asynchronous dummy graph request with a listener that is
     * already disposed when the request is first performed.
     * 
     * <p>
     * The expected outcome is that the request will complete normally, calling
     * execute of the listener procedure once and only once.
     * 
     * <p>
     * The outcome when this test was created was that QueryProcessor would
     * crash internally with a NullPointerException.
     * 
     * This problem is fixed by: https://www.simulationsite.net/trac/simantics/changeset/13869
     * 
     * @throws Exception
     */
	@Test
    public void testDisposedListener() throws Exception{
        final Session session = getSession();

        final Semaphore sem = new Semaphore(0);
        final DataContainer<Throwable> exc = new DataContainer<Throwable>();

        session.asyncRequest(new AsyncRead<Object>() {
            @Override
            public int getFlags() {
                return 0;
            }
		    @Override
		    public int threadHash() {
		    	return hashCode();
		    }
            @Override
            public void perform(AsyncReadGraph graph, AsyncProcedure<Object> procedure) {
                procedure.execute(graph, new Object());
            }
        }, new AsyncListenerAdapter<Object>() {
            @Override
            public boolean isDisposed() {
                return true;
            }
            @Override
            public void exception(AsyncReadGraph graph, Throwable t) {
                System.out.println(getClass().getSimpleName() + ": exception:");
                t.printStackTrace();
                exc.set(t);
                sem.release();
            }
            @Override
            public void execute(AsyncReadGraph graph, Object result) {
                if (DEBUG)
                    System.out.println(getClass().getSimpleName() + ": execute: " + result);
                sem.release();
            }
        });

        sem.acquire();

        Throwable t = exc.get();
        if (t != null) {
            if (t instanceof Exception)
                throw (Exception) t;
            else
                throw new Exception(t);
        }
    }
}
